/* 
 * Main method for Streaming k-means.
 * 
 * This is based on the results in two papers:
 * 
 * It is primarily the implementation of the algorithm from:
 * 
 * 		Vladimir Braverman, Adam Meyerson, Rafail Ostrovsky,
 * 		Alan Roytman, Michael Shindler, and Brian Tagiku.
 * 		"Streaming k-means on Well-Clusterable Data"
 *		From Symposium on Discrete Algorithms, 2011
 * 
 * This implementation was used to produce the following paper, 
 * which improved upon the algorithm and changed some parameters:
 * 
 * 		Adam Meyerson, Michael Shindler, and Alex Wong
 * 		"Fast and Accurate k-means for Large Datasets"
 * 		from Neural Information Processing Systems, 2011
 * 
 * The pieces necessary to run the unaltered algorithm from SODA
 * remain in the code, with comments on how to adjust for use of this.
 * 
 * For practical purposes, we suggest using the implementation with 
 * the parameters set as they are now.
 */


#include <iostream>
#include <math.h>
#include <stdlib.h>
#include <string>
#include <time.h>
#include "Facility.h"
#include "float_random_gen.h"
#include "batch.h"
#include "point.h"

using namespace std;

const bool RESTART_ON_PT = false;

int max(int a, int b)
{
	if( a > b ) return a;
	else return b;
}


/*
 * The following are constant variables.  
 * They are primarily used in the implementation of [BMORST11].
 * In fact, GAMMA has been entirely removed from the implementation
 * A strict implementation of [BMORST11] would use this for the maximum facilities;
 * See [MSW11] for reasons why this is not in use.
 *
 * E is the base of the natural log.  
 *
 * ALPHA is the approximation factor for triangle inequality that the input distance metrics obey.
 *	 Note that for the k-means problem, this is 2.
 *	 For the k-median problem, this is 1.
 *
 * COFL is the approximation factor of the Online Facility Location algorithm's service cost in use.
 * KOFL is a constant from the approximation factor of the Online Facility Location algorithm's facility count.
 *
 * BETA and GAMMA are defined in [BMORST11] and are used primarily in analysis.
 *	BETA is still used for updating facility costs upon phase transition, although may be modified.
 * 	GAMMA is removed from the implementation, although the constant itself is left (in comments)
 *	      so that the user of this code may re-introduce it, should they so choose. 
 */
const float E = 2.718281828;   
	//2.71828182845904523536028747135266249775724709369995...
const float ALPHA = 2.0;
const float COFL = 3 * ALPHA + 2 * ( E / (E - 1) );
const float KOFL = 6 * ALPHA + 1; 
const float BETA = 2 * ALPHA * ALPHA * COFL + 2 * ALPHA;
//const float GAMMA = max( 4 * ALPHA * ALPHA * ALPHA * COFL * COFL + 2 * ALPHA * ALPHA * COFL, BETA * KOFL + 1 );


/*
 * Some of the following auxiliary functions are necessary for the algorithm itself to run.
 * Others are used for parts of the experiment, and remain for anyone who may be interested in it.
 */

/*
 * Makes x a Facility.  Adds it to the list of facilities.  Updates the number of facilities.
 */
void makeFacility(Point * x, Facility *fs[], int & numFacilities, int dim);

/*
 * Determines the k-means cost of using the first k elements of means as Facilities
 * for the first n points in the pointsFile.
 */
double determineKMeansCost(Facility * means[], int k, char * pointsFile, int n, int d);

/*
 * Determines the k-means cost of using the given means for the first n points of the pointsFile.
 * cost1 will contain the cost of using the first k1 Facilities.
 * cost2 will contain the cost of using the first k2 Facilities.
 * 
 * This was used primarily for determining the cost pre- and post- consolidation in experiments.
 */
void determineKMeansCost(Facility * means[], int k1, int k2, char * pointsFile, int n, int d, double & cost1, double & cost2);

/*
 * Determines the cost of using the given k Facilities (means) as a solution for the weighted sketch.
 * This was used primarily in the experiment.
 */
double determineKMeansCostOnSketch(Facility *means[], int k, Facility * sketch[], int n, int d);



int main(int argc, char * argv[]) 
{
	FILE *in;
	//Point *points;
	int length, seed, dimension, k, max_facs, i, j, numOldFacs, numRead, numFacilities;
	char * filename;	
	double cost;
	time_t start, end;
	
	
	if(argc < 6)
	{
		printf("OurAlg.out filename n k d maxFacCount [seed]\n");
		return 1;
	} 
	filename = argv[1];
	length = atoi(argv[2]);
	k = atoi(argv[3]);
	dimension = atoi(argv[4]);
	max_facs = atoi(argv[5]);
	in = fopen(filename,"r");
	float logn = log(length)/log(2);
	
	if( argc == 6 )
	{
		seed = time(0);
	}
	else
	{
		seed = atoi(argv[6]);
	}
	srand(seed);
	
	float delta, f; // delta = cost(x,y)
	Point * x = new Point;
	Facility * y;
	Facility ** facilities;
	facilities = new Facility*[ max_facs ];
	for (i=0; i < max_facs; i++)
	{
		facilities[i] = new Facility;
		init(facilities[i], dimension);

	}
	numFacilities = 0;
	
	init(x, dimension);
	
	time(&start);
	f = 1 / (k*(1+logn));
	getNextPoint(in, x, dimension);

	makeFacility(x, facilities, numFacilities, dimension); // first is a facility.

	numRead = 1;
	
	while( numRead < length )
	{
		getNextPoint(in, x, dimension);
		numRead++;
		y = findNearest( x, facilities, numFacilities, delta );
		if( prob( delta/f) )
		{
			makeFacility(x, facilities, numFacilities, dimension);
			while( numFacilities == max_facs )
			{

				if( RESTART_ON_PT )
				{

					f = f * BETA;
					// clear facilities:
					for (i=0; i < numFacilities; i++)
					{
						freeFacility( facilities[i] );
					}
					numFacilities = 0;
					
					fclose(in);
					in = fopen( filename, "r");
					getNextPoint(in, x, dimension);
					
					makeFacility(x, facilities, numFacilities, dimension);
					numRead = 1;
				}
				else
				{
					f = f * BETA;
					numOldFacs = numFacilities;
					numFacilities = 1;
					// recenter fac 0 at CoM:
					setToCoM( facilities[0] );
					
					for(i=1; i < numOldFacs; i++)
					{
						setToCoM( facilities[ i ]);
						y = findNearest( facilities[i]->p, facilities, numFacilities, delta );

						if( prob(delta/f) )
						{
							// make this a facility
							copyFacility( facilities[numFacilities], facilities[i] );						
							numFacilities++;
							
						}
						else
						{
							assign(y, facilities[i] );
						}
					} // end for
					
					// Need to free memory for everything after the last new facility.
					for (i= numFacilities; i < numOldFacs; i++)
					{
						freeFacility( facilities[i] );
						facilities[i] = new Facility;
						init(facilities[i], dimension);
					}
				} // end if ! RESTART_ON_PT
			} // end "if need to phase transition"
		}
		else
		{
			assign(x,y);
		}
	}

	// now need to consolidate and output k points.
	if( k < numFacilities )
	{
		int * finals = ORSSSeed(facilities, numFacilities, k);

		LocalSearch(facilities, numFacilities, k, finals);

		for (i=0; i < min(numFacilities, k); i++)
		{
			// swap facilities[i] with facilities[ finals[i] ]
			y = facilities[ finals[i] ];
			facilities[ finals[i] ] = facilities[i];
			facilities[i] = y;		
		}
		
#ifdef __DO_FAC_SAMPLING
		double * dists = new double[k];

		// Assign all other facilities to their nearest of the final k.
		// Then facilities[0..k-1] can be used as the final k-means.
		for (i=0; i < k; i++)
		{
			j = findNN(facilities, k, i);
			dists[i] = distanceSQ( facilities[i]->p, facilities[j]->p );
			ballKMeans( facilities[i], dists[i] / 9 );
		}
#endif	
	}
	time(&end);
	fclose(in);
	
	// We are now done processing the file.
	
	// Array facilities[] contains the points.
	// There are min(numFacilities, k) means at the end.
	

	double timeElapsed = difftime(end, start);

	cost = determineKMeansCost(facilities, min(numFacilities,k), filename, length, dimension);

	// Necessary post-algorithm output.
	printf("BMORST11 %s %d %d %.2lf %.2lf %d %d \n",  filename, k, max_facs, timeElapsed, cost, seed, MAX_SAMPLES);

	return 0;
}


void makeFacility(Point * x, Facility *fs[], int & numFacilities, int dim)
{
	copy(fs[ numFacilities ]->p, x);

	// reset total count:
	int i;
	for(i=0; i < x->dimension; i++)
	{
		fs[ numFacilities ]->totals[i] = x->coordinates[i];
	}

	fs[ numFacilities ]->p->weight = 1.0;
	fs[ numFacilities ]->numSamples = 0;
	numFacilities++;	
}

double determineKMeansCost(Facility * means[], int k, char * pointsFile, int n, int d)
{
	FILE * in;
	in = fopen(pointsFile,"r");
	Point * p = new Point;
	init(p, d);
	int i;
	double dis, dis2;
	double cost = 0.0;
	int ptsRead = 0;
	while( ptsRead < n )
	{
		getNextPoint(in, p, d);
		dis = distanceSQ(p, means[0]->p);
		for (i=1; i < k; i++)
		{
			dis2 = distanceSQ(p, means[i]->p);
			if( dis2 < dis )
			{
				dis = dis2;
			}
		}
		cost += dis;
		ptsRead++;
	}
	free(p->coordinates);
	delete p;
	
	return cost;	
}



void determineKMeansCost(Facility * means[], int k1, int k2, char * pointsFile, int n, int d, double & cost1, double & cost2)
{
	cost1 = 0.0;
	cost2 = 0.0;
	FILE * in;
	in = fopen(pointsFile,"r");
	Point * p = new Point;
	init(p, d);
	int i;
	double dis, dis2;
	int ptsRead = 0;
	while( ptsRead < n )
	{
		getNextPoint(in, p, d);
		dis = distanceSQ(p, means[0]->p);
		for (i=1; i < k1; i++)
		{
			dis2 = distanceSQ(p, means[i]->p);
			if( dis2 < dis )
			{
				dis = dis2;
			}
		}
		cost1 += dis;
		
		for(i=k1; i < k2; i++)
		{
			dis2 = distanceSQ(p, means[i]->p);
			if( dis2 < dis )
			{
				dis = dis2;
			}
		}
		cost2 += dis;
		
		
		ptsRead++;
	}
	free(p->coordinates);
	delete p;
}


double distSQToClosest(Facility * pf, Facility * means[], int k)
{
	double closest = distanceSQ( pf->p, means[0]->p );
	double dis;
	for (int i=1; i < k; i++)
	{
		dis = distanceSQ( pf->p, means[i]->p );
		if( dis < closest )
		{
			closest = dis;
		}
	}
	return closest;
}


double determineKMeansCostOnSketch(Facility *means[], int k, Facility * sketch[], int n, int d)
{
	double cost = 0.0;
	double dis;
	for (int i=0; i < n; i++)
	{
		dis = distSQToClosest(sketch[i], means, k);
		dis *= sketch[i]->p->weight;
		cost += dis;			
	}
	return cost;
}
