#include "Facility.h"
#include "point.h"
#include "float_random_gen.h"

using namespace std;

void init(Facility * f, int dim)
{
	f->p = new Point;
	init(f->p, dim);
	f->totals = new float[dim];
	#ifdef __DO_FAC_SAMPLING
	int i;
	for (i=0; i < MAX_SAMPLES; i++)
	{
		f->samples[i] = new Point;
		init( f->samples[i], dim);
	}
	f->numSamples = 0;
	#endif
}


// Frees all auxiliary memory associated with f.
void freeFacility(Facility * f)
{
	freePoint(f->p);
	delete [] f->totals;
	#ifdef __DO_FAC_SAMPLING
	int i;
	for (i=0; i < MAX_SAMPLES; i++)
	{
		freePoint( f->samples[i] );
	}
	#endif
}


Facility * findNearest(Point * p, Facility *fs[], int n, float & distSQ)
{
	int closest = 0;
	int i;
	double dist;
	distSQ = distanceSQ(p, fs[0]->p);
	for (i=1; i < n; i++)
	{
		dist = distanceSQ(p, fs[i]->p);
		if( dist < distSQ)
		{
			distSQ = dist;
			closest = i;
		}
	}
	return fs[closest];
}

void assign(Point * p, Facility * f)
{
	f->p->weight++;
	
	int i;
	for (i=0; i < p->dimension; i++)
	{
		f->totals[i] += p->coordinates[i];
	}
	#ifdef __DO_FAC_SAMPLING
	if( f->numSamples < MAX_SAMPLES)
	{
		copy(f->samples[ f->numSamples ], p);
		f->numSamples++;
	}
	else
	{
		for (i=0; i < MAX_SAMPLES; i++)
		{
			if( prob(1.0/(MAX_SAMPLES + 1)))
			{
				copy(f->samples[i], p);
				return;
			}
		}
	}
	#endif
}

#ifdef __DO_FAC_SAMPLING

int numToKeepFromA(int n1, int n2)
{
	int num = 0;
	int sum = n1+n2;
	int i;
	for (i=0; i < MAX_SAMPLES; i++)
	{
		if( prob( 1.0* n1 / sum ) )
		{
			num++;
			n1--;
			sum--;
		}
		else
		{
			n2--;
			sum--;
		}
	}
	return num;
}


void mergeSamples(Facility * a, Facility * b)
{
	int i, n1, n2;
	n1 = a->p->weight;
	n2 = b->p->weight;
	
	// Note:  If n1 >= MAX_SAMPLES, samples full.  
	// 			Not full otherwise.
	// 			Same with n2.  
	
	if( n1 + n2 <= MAX_SAMPLES )
	{
		// add all samples from b to a.
		for(i=0; i < n2; i++)
		{
			// move b->samples[i] to a->samples[i+ n1]
			copy( a->samples[i+n1], b->samples[i] );
		}
		 a->numSamples = n1 + n2;		
		// update a's total.
		
		return;	// done
	}
	
	// else, we have at least MAX_SAMPLES total.
	// Need to determine how many from a to keep
	// (and, similarly, how many from b to keep)
	int keepFromA = numToKeepFromA(n1, n2);
	int keepFromB = MAX_SAMPLES - keepFromA;
	for(i=0; i < keepFromB; i++)
	{
		// move b->samples[i] to a->samples[i + keepFromA].
		copy( a->samples[i + keepFromA], b->samples[i] );
	}
	
	// NOTE:  Do *NOT* delete anything in b.
	// 		  That is reserved for main (or equivalent)
	
}
#endif

void assign(Facility * a, Facility * b)
{
#ifdef __DO_FAC_SAMPLING
	mergeSamples(a,b);
#endif
	a->p->weight += b->p->weight;
	int i;
	for (i=0; i < a->p->dimension; i++)
	{
		a->totals[i] += b->totals[i];
	}
}


// recenters f at its center of mass of all assigned points
void setToCoM(Facility * f)
{
	int i;
	for (i=0; i < f->p->dimension;  i++)
	{
		f->p->coordinates[i] = f->totals[i] / f->p->weight;
	}
}


int findNN(Facility * fs[], int n, int x)
{
	double dis, dis2; 
	int j, nn;
	if( x == 0 )
	{
		dis = distanceSQ( fs[0]->p, fs[1]->p );
		nn = 1;
	}
	else
	{
		dis = distanceSQ( fs[0]->p, fs[x]->p );
		nn = 0;
	}
	for (j=1; j < n; j++)
	{
		if ( j != x )
		{
			dis2 = distanceSQ( fs[x]->p, fs[j]->p);
			if( dis2 < dis )
			{
				nn = j;
				dis = dis2;
			}
		}
	}
	return nn;
}


void copyFacility(Facility * a, Facility * b)
{
	int i;
	copy(a->p, b->p);
	for (i=0; i < a->p->dimension; i++)
	{
		a->totals[i] = b->totals[i];
	}
	
	for (i=0; i < MAX_SAMPLES; i++)
	{
		copy(a->samples[i], b->samples[i] );
	}
	a->numSamples = b->numSamples;
}

void ballKMeans(Facility * f, double dist)
{
	// recenter f at the CoM of all sampled points within the given distance.
	
	int numInBall = 1;
	int dim = f->p->dimension;
	float * totals = (float*) calloc(dim, sizeof(float)); 
	int i, j;
	for (i=0; i < dim; i++)
	{
		totals[i] = f->p->coordinates[i];
	}

	for (i=0; i < f->numSamples; i++)
	{
		// if f->samples[i] within dist of f->p
		// add it to the balled coordinates.
		if( distanceSQ( f->samples[i], f->p ) <= dist )
		{
			for (j=0; j < dim; j++)
			{
				totals[j] += f->samples[i]->coordinates[j];
			}
			numInBall++;
		}
	}

	for (i=0; i < dim; i++)
	{
		f->p->coordinates[i] = totals[i] / numInBall;
	}
	
	free(totals);
}
