/*
 * A structure to represent a Facility in Streaming k-means.
 * 
 * If sampling is to be used (as per the rules in [BMORST11], 
 *	ensure that  __DO_FAC_SAMPLING is defined
 *	(get rid of the definition if no sampling is to be used)
 *
 *	Set MAX_SAMPLES to be the number of samples each Facility 
 *	will maintain, at most.  
 *	Within the context of [BMORST11], this means that
 *		epsilon = 1/MAX_SAMPLES
 */

#ifndef __FACILITY_H
#define __FACILITY_H

#define __DO_FAC_SAMPLING
// If undefined, this will not do any sampling.

struct Point;

#ifdef __DO_FAC_SAMPLING
const int MAX_SAMPLES = 4;
#endif

/*
 * For reasons I won't get into here, we made the decision to
 * not use polymorphism to represent Facilities, even though
 * a Facility is-a Point, with additional properties.
 */
struct Facility
{
	Point * p; // includes dim, weight
	float * totals; // for use in computing CoM
	
	#ifdef __DO_FAC_SAMPLING
	// SAMPLING:
	Point * samples[MAX_SAMPLES];
	int numSamples;
	#endif
};

/*
 * initializes an empty facility, for a point with the given dimensions.
 */
void init(Facility * f, int dim);

/*
 * Makes a deep copy of a = b.
 */
void copyFacility(Facility * a, Facility * b);

/*
 * Returns a pointer to the nearest facility to p from the given list.
 * The square of the distance between the two point and the nearest facility
 * is returned via the reference parameter distSQ.
 */ 
Facility * findNearest(Point * p, Facility *fs[], int n, float & distSQ);

/*
 * Assigns the given point p to the designated facility.
 */
void assign(Point * p, Facility * f);

/*
 * Assigns facility b to a.
 * 	The samples are merged, as per [BMORST11]
 */ 
void assign(Facility * a, Facility * b); 

/*
 * Releases the memory associated with this facility.
 */
void freeFacility(Facility * f); 

/*
 * Sets this Facility's designation point to be the 
 * center of mass of all the points assigned to this
 * facility (samples or otherwise).
 */
void setToCoM(Facility * f);

/*
 * Sets this Facility's designation point to be the 
 * center of mass of all sampled points (including the
 * designation point itself) within the given distance
 * of the designation point.
 * 
 * Note that this is the "ball k-means" step that is 
 * used in [ORSS06] and [BMORST11].
 * Typically, this distance is one-third of the distance
 * from f's designation point to the nearest other 
 * semi-final Facility's designation point, and this
 * replaces the Lloyd-descent step in [ORSS06]
 */
void ballKMeans(Facility * f, double dist);

/*
 * Finds the nearest neighbor of fs[i].
 * That is, it finds the element of fs[] - fs[i]
 *		that is closest to fs[i] without being fs[i]
 */
int findNN(Facility * fs[], int n, int i);

#endif
