/* 
 * Main method for Streaming k-means with approximate nearest neighbor.
 * 
 * This is based on the results in two papers:
 * 
 * It started as the implementation of the algorithm from:
 * 
 * 		Vladimir Braverman, Adam Meyerson, Rafail Ostrovsky,
 * 		Alan Roytman, Michael Shindler, and Brian Tagiku.
 * 		"Streaming k-means on Well-Clusterable Data"
 *		From Symposium on Discrete Algorithms, 2011
 * 
 * This implementation was used to produce the following paper, 
 * which improved upon the algorithm and changed some parameters:
 * 
 * 		Adam Meyerson, Michael Shindler, and Alex Wong
 * 		"Fast and Accurate k-means for Large Datasets"
 * 		from Neural Information Processing Systems, 2011
 * 
 *
 * In particular, it is the observation from the second paper that 
 * a large fraction of the running time is spent looking for the 
 * exact nearest neighbor of each new point.
 *
 * This main method will, instead, use a very simple approximate
 * nearest neighbor technique to find the Facility to use for
 * each successive point.
 */


#include <math.h>
#include <iostream>
#include <stdlib.h>
#include <string>
#include <time.h>
#include "Facility.h"
#include "float_random_gen.h"
#include "point.h"
#include "batch.h"

using namespace std;

const bool RESTART_ON_PT = false;

int max(int a, int b)
{
	if( a > b ) return a;
	else return b;
}

struct ApproximateFacility
{
	int facilityNumber; // where is the facility this represents stored?
	double projection; // Where is this projected, along the approximate facility set?
};



/*
 * For a description of the constants, see BMORST11.cpp
 */

const float E = 2.718281828;   
	//2.71828182845904523536028747135266249775724709369995...
const float ALPHA = 2.0;
const float COFL = 3 * ALPHA + 2 * ( E / (E - 1) );
const float KOFL = 6 * ALPHA + 1; 
const float BETA = 2 * ALPHA * ALPHA * COFL + 2 * ALPHA;
//const float GAMMA = max( 4 * ALPHA * ALPHA * ALPHA * COFL * COFL + 2 * ALPHA * ALPHA * COFL, BETA * KOFL + 1 );



void makeFacility(Point * x, Facility *fs[], ApproximateFacility * afs, int & numFacilities, int dim, float * proj);

Facility * findApproxNearest(Point * p, Facility *fs[], ApproximateFacility * afs, int n, float & distSQ, float * proj, int dim);

double determineKMeansCost(Facility * means[], int k, char * pointsFile, int n, int d);
void determineKMeansCost(Facility * means[], int k1, int k2, char * pointsFile, int n, int d, double & cost1, double & cost2);

float project(float x[], float y[], int d);
void insert(ApproximateFacility * afs, int n); // inserts the nth one so that they're sorted by project #.
int binarySearch(ApproximateFacility * afs, int n, double target);



int main(int argc, char * argv[]) 
{
	FILE *in;
	//Point *points;
	int length, seed, dimension, k, max_facs, i, j, numOldFacs, numRead, numFacilities;
	char * filename;	
	double cost;
	time_t start, end;
	
	if(argc < 6)
	{
		printf("SWM11 filename n k d maxFacCount [seed]\n");
		return 1;
	} 
	filename = argv[1];
	length = atoi(argv[2]);
	k = atoi(argv[3]);
	dimension = atoi(argv[4]);
	max_facs = atoi(argv[5]);
	in = fopen(filename,"r");
	float logn = log(length)/log(2);
	
	if( argc == 6 )
	{
		seed = time(0);
	}
	else
	{
		seed = atoi(argv[6]);
	}
	srand(seed);
	
	float delta, f; // delta = cost(x,y)
	Point * x = new Point;
	Facility * y;
	Facility ** facilities;
	facilities = new Facility*[ max_facs ];
	ApproximateFacility * approxFacs = new ApproximateFacility[ max_facs ];
	float * projectedPoint = new float[ dimension ];
	
	// Fill projected point, each coord[i] with random [0,1).
	for (i=0; i < dimension; i++)
	{
		projectedPoint[i] = randfloat();
	}
	
	for (i=0; i < max_facs; i++)
	{
		facilities[i] = new Facility;
		init(facilities[i], dimension);
	}
	numFacilities = 0;
	
	init(x, dimension);
	
	time(&start);
	f = 1 / (k*(1+logn));
	getNextPoint(in, x, dimension);

	makeFacility(x, facilities, approxFacs, numFacilities, dimension, projectedPoint); // first is a facility.
	

	numRead = 1;
	
	while( numRead < length )
	{
		getNextPoint(in, x, dimension);
		numRead++;

		y = findApproxNearest( x, facilities, approxFacs, numFacilities, delta, projectedPoint, dimension ); 


		if( prob( delta/f) )
		{
			makeFacility(x, facilities, approxFacs, numFacilities, dimension, projectedPoint);
			while( numFacilities == max_facs )
			{
				// phase transition:
				f = f * BETA;
				numOldFacs = numFacilities;
				numFacilities = 1;
				// recenter fac 0 at CoM:
				setToCoM( facilities[0] );
				// reproject it.
				approxFacs[0].projection = project(facilities[0]->p->coordinates, projectedPoint, dimension );
				approxFacs[0].facilityNumber = 0;
				
				for(i=1; i < numOldFacs; i++)
				{
					setToCoM( facilities[ i ]);
				
					y = findApproxNearest( facilities[i]->p, facilities, approxFacs, numFacilities, delta, projectedPoint, dimension ); 
					if( prob(delta/f) )
					{
						// make this a facility
						// Do so by moving the memory to the earlier spot.							
						copyFacility( facilities[numFacilities], facilities[i] );
						approxFacs[numFacilities].projection = project(facilities[numFacilities]->p->coordinates, projectedPoint, dimension );
						approxFacs[numFacilities].facilityNumber = numFacilities;
						insert(approxFacs, numFacilities);
						
						numFacilities++;
					}
					else
					{
						assign(y, facilities[i]);
					}
				} // end for
			} // end "if need to phase transition"
		}
		else
		{
			assign(x,y);
		}
	}
	
	// now need to consolidate and output k points.
	if( k < numFacilities )
	{
		int * finals = ORSSSeed(facilities, numFacilities, k);

		LocalSearch(facilities, numFacilities, k, finals);

		for (i=0; i < min(numFacilities, k); i++)
		{
			// swap facilities[i] with facilities[ finals[i] ]
			y = facilities[ finals[i] ];
			facilities[ finals[i] ] = facilities[i];
			facilities[i] = y;		
		}
		
#ifdef __DO_FAC_SAMPLING
		double * dists = new double[k];

		// Assign all other facilities to their nearest of the final k.
		// Then facilities[0..k-1] can be used as the final k-means.
		for (i=0; i < k; i++)
		{
			j = findNN(facilities, k, i);
			dists[i] = distanceSQ( facilities[i]->p, facilities[j]->p );
			ballKMeans( facilities[i], dists[i] / 9 );
		}
#endif	
	}
	time(&end);
	fclose(in);

	double timeElapsed = difftime(end, start);

	cost = determineKMeansCost(facilities, min(numFacilities,k), filename, length, dimension);

	// Necessary post-algorithm output.
	printf("SWM11 %s %d %d %.2lf %.2lf %d %d\n", filename, k, max_facs, timeElapsed, cost, seed, MAX_SAMPLES);
	
// clean up:
	delete x; 
	for (i=0; i < max_facs; i++)
	{
			freeFacility( facilities[i] );
	}
	delete [] facilities;
	delete [] approxFacs; 
	delete [] projectedPoint; 


	return 0;
}


void makeFacility(Point * x, Facility *fs[], ApproximateFacility * afs, int & numFacilities, int dim, float * proj)
{
	// remember to copy x, because we will change it later.		
	copy(fs[ numFacilities ]->p, x);
	
#ifdef __DO_FAC_SAMPLING
	fs[ numFacilities ]->numSamples = 0;
#endif

	fs[ numFacilities ]->p->weight = 1.0;
	int i;
	for (i=0; i < x->dimension; i++)
	{
		fs[ numFacilities]->totals[i] = x->coordinates[i];
	}
	
	afs[numFacilities].facilityNumber = numFacilities;
	afs[numFacilities].projection = project(x->coordinates, proj, dim );
	
	insert( afs, numFacilities );

	numFacilities++;	
}

double determineKMeansCost(Facility * means[], int k, char * pointsFile, int n, int d)
{
	FILE * in;
	in = fopen(pointsFile,"r");
	Point * p = new Point;
	init(p, d);
	int i;
	double dis, dis2;
	double cost = 0.0;
	int ptsRead = 0;
	while( ptsRead < n )
	{
		getNextPoint(in, p, d);
		dis = distanceSQ(p, means[0]->p);
		for (i=1; i < k; i++)
		{
			dis2 = distanceSQ(p, means[i]->p);
			if( dis2 < dis )
			{
				dis = dis2;
			}
		}
		cost += dis;
		ptsRead++;
	}
	free(p->coordinates);
	delete p;
	
	return cost;	
}



void determineKMeansCost(Facility * means[], int k1, int k2, char * pointsFile, int n, int d, double & cost1, double & cost2)
{
	cost1 = 0.0;
	cost2 = 0.0;
	FILE * in;
	in = fopen(pointsFile,"r");
	Point * p = new Point;
	init(p, d);
	int i;
	double dis, dis2;
	int ptsRead = 0;
	while( ptsRead < n )
	{
		getNextPoint(in, p, d);
		dis = distanceSQ(p, means[0]->p);
		for (i=1; i < k1; i++)
		{
			dis2 = distanceSQ(p, means[i]->p);
			if( dis2 < dis )
			{
				dis = dis2;
			}
		}
		cost1 += dis;
		
		for(i=k1; i < k2; i++)
		{
			dis2 = distanceSQ(p, means[i]->p);
			if( dis2 < dis )
			{
				dis = dis2;
			}
		}
		cost2 += dis;
		
		
		ptsRead++;
	}
	free(p->coordinates);
	delete p;
}


// returns the dot product of x and y.
float project(float x[], float y[], int d)
{
	float sum = 0.0;
	int i;
	for (i=0; i < d; i++)
	{
		sum += ( x[i] * y[i] );
	}
	return sum;
}

/*
 * After afs[n] has become an ApproxFacility, this function will
 * insert it into the array so that the array remains
 * in ascending (non-decreasing) order.
 */
void insert(ApproximateFacility * afs, int n)
{
	int i = n;
	int tempint;
	double tempdbl;
	while( i > 0 && afs[i-1].projection > afs[i].projection )
	{
		// swap i-1, i within afs.
		tempint = afs[i-1].facilityNumber;
		afs[i-1].facilityNumber = afs[i].facilityNumber;
		afs[i].facilityNumber = tempint;
		
		tempdbl = afs[i-1].projection;
		afs[i-1].projection = afs[i].projection;
		afs[i].projection = tempdbl;
		
		i--;
	}
}

Facility * findApproxNearest(Point * p, Facility *fs[], ApproximateFacility * afs, int n, float & distSQ, float * proj, int dim)
{
	float projection = project(p->coordinates, proj, dim);
	int loc = binarySearch(afs, n, projection);
	
	if( loc == -1 )
	{
		// nearest is fs[ afs[0].facilityNumber ].
		distSQ = distanceSQ( p, fs[ afs[0].facilityNumber]->p );
		return fs[ afs[0].facilityNumber ];
	}
	else if ( loc == n-1)
	{
		// nearest is fs[ afs[n-1].facilityNumber ].
		distSQ = distanceSQ( p, fs[ afs[n-1].facilityNumber]->p );
		return fs[ afs[n-1].facilityNumber ];
	}
//	else:
	// either fs[ afs[loc].facilityNumber ]
	// or 	  fs[ afs[loc+1].facilityNumber ]
	distSQ = distanceSQ( p, fs[ afs[loc].facilityNumber]->p);
	double dis = distanceSQ( p, fs[ afs[loc+1].facilityNumber]->p );
	if( distSQ <= dis )
	{
		return fs[afs[loc].facilityNumber];
	}
	else
	{
		distSQ = dis;
		return fs[ afs[loc+1].facilityNumber ];
	}	
}

// finds where target lies within the given array.
// Returns the largest i such that nums[i] <= target.
// 	If target < nums[0], will return -1. 
int binarySearch(ApproximateFacility * afs, int n, double target)
{
	if( target < afs[0].projection ) 
		return -1;
	if ( target > afs[ n-1 ].projection )
		return n-1;
	int low = 0;
	int high = n-1;
	int mid;
	// Invariant:  nums[low] <= target <= nums[high]
	while( high - low > 1 )
	{
		mid = (high + low) / 2; 
		if ( target <= afs[mid ].projection )
		{
			high = mid;
		}
		if ( afs[mid].projection <= target )
		{
			low = mid;
		}
	}
	return low;
}
