#include "batch.h"
#include "point.h"
#include "Facility.h"
#include "float_random_gen.h"
#include <iostream> // for cerr

using namespace std;


// Fills the first two of ctrs according to the ORSS distribution.
void fillFirstTwo(Facility * fs[], int numFacilities, int * ctrs, int dim, double dist_sq[]);
float distanceSQ(float a[], float b[], int dim);
int selectWithProb(double probs[], int n);

/*
 * Fills the array probs[*] in such a way that:
 * 		- The sum of the probabilities is 1.
 * 		- probs[i] is proportional to dist_sq[i]
 */
void determineProbabilities(double dist_sq[], Facility * fs[], double probs[], int n);


void updateDistSQ(double dist_sq[], int n, int i, Facility * fs[]);

bool contains(int a[], int n, int x);
double kMeansCost(Facility * fs[], int n, int k, int start[]);



/*
 * Returns an array of integers of size k
 * These are each distinct and in the range [0, n-1]
 * These represent k initial means for a k-means solution
 * 	as per the ORSS distribution 
 */
int * ORSSSeed(Facility * fs[], int n, int k)
{
	int dim = fs[0]->p->dimension;
	int * ctrs = new int[k];
	int i;
	double * probs  = new double[ n ];
	double * dist_sq = new double[ n ];
	// compute these among first two.
	fillFirstTwo(fs, n, ctrs, dim, dist_sq);
	for (i=2; i < min(n,k); i++)
	{
		// select ith point:
		determineProbabilities(dist_sq, fs, probs, n);
		ctrs[i] = selectWithProb(probs, n);
		
		updateDistSQ(dist_sq, n, ctrs[i], fs);
		
	}
	delete [] dist_sq;
	delete [] probs;
	return ctrs;
}

// assumes (and this is OK) that ctrs can hold at least ctrs[0] & ctrs[1]
// When this returns, dist_sq must hold the right values
void fillFirstTwo(Facility * fs[], int n, int * ctrs, int dim, double dist_sq[])
{
	double OPT_1 = 0; // optimal 1-means cost.
	float * com = new float[ dim ]; // weighted com of points
	double * probs = new double[ n ];
	double sum = 0.0;
	
	int total_weight = 0; 
	int i, j, weight;
	double ctr_to_c1_sq;

	for (i=0; i < n; i++)
	{
		weight = fs[i]->p->weight;
		total_weight += weight;
		for(j=0; j < dim; j++)
		{
			com[j] += ( weight * fs[i]->p->coordinates[j] );
		}
	}
	for (j=0; j < dim; j++)
	{
		com[j] = com[j] / total_weight;
	}
	for (i=0; i < n; i++)
	{
		dist_sq[i] = distanceSQ( com, fs[i]->p->coordinates, dim);
		OPT_1 += dist_sq[i] * fs[i]->p->weight;
	}
	for (i=0; i < n; i++)
	{
		probs[i] = (OPT_1 + total_weight* dist_sq[i] )/ (2 * total_weight * OPT_1 );
		sum += probs[i];
	}
	
	for (i=0; i < n; i++)
	{
		probs[i] = probs[i] / sum;
	}		

	// Select first with probability equal to probs[i].
	ctrs[0] = selectWithProb(probs, n);
	if( ctrs[0] == n)
	{
		cerr << "Error in ORSS Seed selection: first point." << endl;
		exit(1);
	} 
	ctr_to_c1_sq = distanceSQ(com, fs[ ctrs[0] ]->p->coordinates, dim);
	Facility * y = fs[ ctrs[0] ]; // first center
	
	// refill dist_sq and probs for point two:
	sum = 0.0; 
	probs[ ctrs[0] ] = 0;
	for (i=0; i < n; i++)
	{
		if( i != ctrs[0] )
		{
			dist_sq[i] = distanceSQ( fs[i]->p, y->p ) * fs[i]->p->weight;
			probs[i] = dist_sq[i] / ( OPT_1 + total_weight * ctr_to_c1_sq );
			sum += probs[i];
		}
	}
	for (i=0; i < n; i++)
	{
		probs[i] = probs[i] / sum;
	}		
	
	ctrs[1] = selectWithProb(probs, n);

	dist_sq[ ctrs[1] ] = 0;

	
	if( ctrs[1] == n)
	{
		cerr << "Error in ORSS Seed selection: second point." << endl;
		exit(1);
	}
	
	for (i=0; i < n; i++)
	{
		dist_sq[i] = min( distanceSQ(fs[i]->p, fs[ctrs[0]]->p), 
						  distanceSQ(fs[i]->p, fs[ctrs[1]]->p ) );
	}
	
	delete [] com;
	delete [] probs;
}

// Given that sum of probabilities = 1, select i [ 0, n-1] w/ prob p[i]
// If the sum is not 1, the guarantee of probabilities does not hold.
int selectWithProb(double probs[], int n)
{
	double cutoff = randfloat(); // is [0,1]
	double cur_pt = 0.0;
	int i;
	for (i=0; i < n; i++)
	{
		cur_pt += probs[i];
		if ( cur_pt >= cutoff )
		{
			return i;
		}
	}
	
	cout << "Going to return n = " << n << endl;
	cout << "cutoff = " << cutoff << endl;
	double sum = 0.0;
	for (i=0; i < n; i++)
	{
		sum += probs[i];
		cout << " prob[ " << i << " ] = " << probs[i] << endl;
	}
	cout << "sum = " << sum << endl;
	
	return n;
}


// returns the square of the distance between dim-dimensional
// points, given only their coordinates
float distanceSQ(float a[], float b[], int dim)
{
	float sum = 0.0;
	int i;
	float it;
	for (i=0; i < dim; i++)
	{
		it = a[ i ] - b[i];
		it = it * it;
		sum += it;
	}
	if( sum == 0 )
	{
		cerr << "Returning 0 for sum of distanceSQ between arrays of floats." << endl;
		cerr << "a = ";
		for (i=0; i < dim; i++)
		{
			cerr << ", " << a[i];
		}
		cerr << endl << " b = ";
		for (i=0; i < dim; i++)
		{
			cerr << ", " << b[i];
		}		
		cerr << endl;	
	}
	return sum;	
}

void updateDistSQ(double dist_sq[], int n, int chosen, Facility * fs[])
{
	int i;
	for (i=0; i < n; i++)
	{
		dist_sq[i] = min( dist_sq[i], distanceSQ( fs[i]->p, fs[ chosen ]->p ));
	}
}

void determineProbabilities(double dist_sq[], Facility * fs[], double probs[], int n)
{
	double sum = 0;
	int i;
	for (i=0; i < n; i++)
	{
		probs[i] = dist_sq[i] * fs[i]->p->weight;
		sum += probs[i];
	}
	//cerr << "\tSum in detProb = " << sum << endl;
	for (i=0; i < n; i++)
	{
		probs[i] = probs[i] / sum;
	}
}

/*
 * Performs one-swap local search for k-means on the given set of facilities.
 * The starting point for the local search is in start[].
 * 		{ fs[start[0]], fs[start[1]], ... fs[start[k-1]] }
 * The ending point is returned via the start[] parameter.
 */
void LocalSearch(Facility * fs[], int n, int k, int start[])
{
	int i, ii, j, was;
	bool changed;
	double bestCost = kMeansCost(fs, n, k, start);
	double cost;
	int * best = new int[k];
	do
	{
		changed = false;
		for (i=0; i < n; i++)
		{
			if( ! contains( start, k, i ) )
			{
				for (j=0; j < k; j++)
				{
					was = start[j];
					start[j] = i;
					cost = kMeansCost(fs, n, k, start);
					if( cost < bestCost )
					{
					
						// copy start into best
						for (ii = 0; ii < k; ii++)
						{
							best[ii] = start[ii];
						}
						bestCost = cost;
						changed = true;
					}
					start[j] = was;
				}
			}
		}
		if( changed )
		{
			for (i=0; i < k; i++)
			{
				start[i] = best[i];
			}
		}
	}while( changed );

	delete [] best;
}

bool contains(int a[], int n, int x)
{
	// is there an i s.t. a[i] == x?
	int i;
	for (i=0; i < n; i++)
	{
		if ( a[i] == x ) return true;
	}
	
	return false;
}

// returns the distance from fs[x] to the nearest of fs[ start[0] ] etc
double distNearest(Facility *fs[], int x, int k, int start[])
{
	double dis, dis2;
	int i;
	dis = distanceSQ( fs[x]->p, fs[ start[0] ]->p );
	for (i=1; i < k; i++)
	{
		dis2 = distanceSQ( fs[x]->p, fs[ start[i] ]->p );
		if( dis2 < dis)
		{
			dis = dis2;
		}
	}
	return dis;
}

/*
 *  returns the k-means cost of using the given facilities
 * 		{ fs[start[0]], fs[start[1]], ... fs[start[k-1]] }
 *
 */
double kMeansCost(Facility * fs[], int n, int k, int start[])
{
	double sum = 0.0;
	int i;
	for (i=0; i < n; i++)
	{
		sum += distNearest(fs, i, k, start);
	}
	return sum;
}
